/*
 * Copyright 2016 - 2025 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */


#include <ox/claw/read.hpp>
#include <ox/std/algorithm.hpp>
#include <ox/std/buffer.hpp>
#include <ox/std/memory.hpp>

#include <turbine/clipboard.hpp>
#include <keel/media.hpp>

#include <nostalgia/gfx/ptidxconv.hpp>

#include "commands/commands.hpp"
#include "commands/addsubsheetcommand.hpp"
#include "commands/cutpastecommand.hpp"
#include "commands/deletetilescommand.hpp"
#include "commands/drawcommand.hpp"
#include "commands/inserttilescommand.hpp"
#include "commands/palettechangecommand.hpp"
#include "commands/rmsubsheetcommand.hpp"
#include "commands/updatesubsheetcommand.hpp"
#include "tilesheeteditormodel.hpp"

namespace nostalgia::gfx {

// delete pixels of all non-leaf nodes
static void normalizeSubsheets(TileSheet::SubSheet &ss) noexcept {
	if (ss.subsheets.empty()) {
		for (auto &child : ss.subsheets) {
			normalizeSubsheets(child);
		}
	} else {
		ss.pixels.clear();
	}
}


Palette const TileSheetEditorModel::s_defaultPalette = {
	.colorNames = {ox::Vector<ox::String>{{}}},
	.pages = {{"Page 1", ox::Vector<Color16>(128)}},
};

TileSheetEditorModel::TileSheetEditorModel(
		studio::StudioContext &sctx, ox::StringViewCR path, studio::UndoStack &undoStack):
	m_sctx(sctx),
	m_tctx(m_sctx.tctx),
	m_path(path),
	m_img(*readObj<TileSheet>(keelCtx(m_tctx), m_path).unwrapThrow()),
	// ignore failure to load palette
	m_pal(readObj<Palette>(keelCtx(m_tctx), m_img.defaultPalette).value),
	m_undoStack(undoStack) {
	normalizeSubsheets(m_img.subsheet);
	m_pal.updated.connect(this, &TileSheetEditorModel::markUpdated);
	m_undoStack.changeTriggered.connect(this, &TileSheetEditorModel::markUpdatedCmdId);
}

void TileSheetEditorModel::cut() {
	if (!m_selection) {
		return;
	}
	TileSheetClipboard blankCb;
	auto cb = ox::make_unique<TileSheetClipboard>();
	auto const&s = activeSubSheet();
	iterateSelectionRows(*m_selection, [&](int const x, int const y) {
		auto pt = ox::Point{x, y};
		auto const idx = gfx::idx(s, pt);
		auto const c = getPixel(s, idx);
		pt -= m_selection->a;
		cb->addPixel(pt, c);
		blankCb.addPixel(pt, 0);
	});
	auto const pt1 = m_selection->a;
	auto const pt2 = ox::Point{s.columns * TileWidth, s.rows * TileHeight};
	turbine::setClipboardObject(m_tctx, std::move(cb));
	pushCommand(ox::make<CutPasteCommand>(
		CommandId::Cut, m_img, m_activeSubsSheetIdx, pt1, pt2, blankCb));
}

void TileSheetEditorModel::copy() {
	if (!m_selection) {
		return;
	}
	auto cb = ox::make_unique<TileSheetClipboard>();
	iterateSelectionRows(*m_selection, [&](int const x, int const y) {
		auto pt = ox::Point{x, y};
		const auto&s = activeSubSheet();
		const auto idx = gfx::idx(s, pt);
		const auto c = getPixel(s, idx);
		pt -= m_selection->a;
		cb->addPixel(pt, c);
	});
	turbine::setClipboardObject(m_tctx, std::move(cb));
}

void TileSheetEditorModel::paste() {
	if (!m_selection) {
		return;
	}
	auto [cb, err] = turbine::getClipboardObject<TileSheetClipboard>(m_tctx);
	if (err) {
		oxLogError(err);
		oxErrf("Could not read clipboard: {}", toStr(err));
		return;
	}
	auto const&s = activeSubSheet();
	auto const pt1 = m_selection->a;
	auto const pt2 = ox::Point{s.columns * TileWidth, s.rows * TileHeight};
	pushCommand(ox::make<CutPasteCommand>(
		CommandId::Paste, m_img, m_activeSubsSheetIdx, pt1, pt2, *cb));
}

bool TileSheetEditorModel::acceptsClipboardPayload() const noexcept {
	auto const cb = getClipboardObject<TileSheetClipboard>(m_tctx);
	return cb.ok();
}

ox::StringView TileSheetEditorModel::palPath() const noexcept {
	auto &path = m_img.defaultPalette;
	constexpr ox::StringView uuidPrefix = "uuid://";
	if (ox::beginsWith(path, uuidPrefix)) {
		auto const uuid = substr(path, uuidPrefix.bytes());
		auto const out = keelCtx(m_tctx).uuidToPath.at(uuid);
		if (out.error) {
			return {};
		}
		return *out.value;
	} else {
		return path;
	}
}

ox::Error TileSheetEditorModel::setPalette(ox::StringViewCR path) noexcept {
	OX_REQUIRE(uuid, keelCtx(m_tctx).pathToUuid.at(path));
	pushCommand(ox::make<PaletteChangeCommand>(
		activeSubSheetIdx(), m_img, uuid->toString()));
	return {};
}

void TileSheetEditorModel::setPalettePage(size_t const pg) noexcept {
	m_palettePage = ox::clamp<size_t>(pg, 0, m_pal->pages.size() - 1);
	m_updated = true;
}

size_t TileSheetEditorModel::palettePage() const noexcept {
	return m_palettePage;
}

void TileSheetEditorModel::drawCommand(ox::Point const&pt, std::size_t const palIdx) noexcept {
	const auto &activeSubSheet = getSubSheet(m_img, m_activeSubsSheetIdx);
	if (pt.x >= activeSubSheet.columns * TileWidth || pt.y >= activeSubSheet.rows * TileHeight) {
		return;
	}
	const auto idx = gfx::idx(activeSubSheet, pt);
	if (m_ongoingDrawCommand) {
		m_updated = m_updated || m_ongoingDrawCommand->append(idx);
	} else if (getPixel(activeSubSheet, idx) != palIdx) {
		pushCommand(ox::make<DrawCommand>(
			m_img, m_activeSubsSheetIdx, idx, static_cast<int>(palIdx)));
	}
}

void TileSheetEditorModel::endDrawCommand() noexcept {
	m_ongoingDrawCommand = nullptr;
}

void TileSheetEditorModel::addSubsheet(TileSheet::SubSheetIdx const&parentIdx) noexcept {
	pushCommand(ox::make<AddSubSheetCommand>(m_img, parentIdx));
}

void TileSheetEditorModel::rmSubsheet(TileSheet::SubSheetIdx const&idx) noexcept {
	pushCommand(ox::make<RmSubSheetCommand>(m_img, idx));
}

void TileSheetEditorModel::insertTiles(
	TileSheet::SubSheetIdx const&idx, std::size_t const tileIdx, std::size_t const tileCnt) noexcept {
	pushCommand(ox::make<InsertTilesCommand>(m_img, idx, tileIdx, tileCnt));
}

void TileSheetEditorModel::deleteTiles(
	TileSheet::SubSheetIdx const&idx, std::size_t const tileIdx, std::size_t const tileCnt) noexcept {
	pushCommand(ox::make<DeleteTilesCommand>(m_img, idx, tileIdx, tileCnt));
}

ox::Error TileSheetEditorModel::updateSubsheet(
	TileSheet::SubSheetIdx const&idx, ox::StringViewCR name, int const cols, int const rows) noexcept {
	OX_REQUIRE(cmd, ox::makeCatch<UpdateSubSheetCommand>(m_img, idx, name, cols, rows));
	pushCommand(cmd);
	return {};
}

void TileSheetEditorModel::setActiveSubsheet(TileSheet::SubSheetIdx const&idx) noexcept {
	m_activeSubsSheetIdx = idx;
	this->activeSubsheetChanged.emit(m_activeSubsSheetIdx);
}

void TileSheetEditorModel::fill(ox::Point const&pt, int const palIdx) noexcept {
	auto const&activeSubSheet = getSubSheet(m_img, m_activeSubsSheetIdx);
	// build idx list
	if (pt.x >= activeSubSheet.columns * TileWidth || pt.y >= activeSubSheet.rows * TileHeight) {
		return;
	}
	ox::Array<bool, PixelsPerTile> updateMap = {};
	auto const oldColor = getPixel(activeSubSheet, pt);
	getFillPixels(activeSubSheet, updateMap, pt, oldColor);
	ox::Vector<std::size_t> idxList;
	auto i = gfx::idx(activeSubSheet, pt) / PixelsPerTile * PixelsPerTile;
	for (auto const u : updateMap) {
		if (u) {
			idxList.emplace_back(i);
		}
		++i;
	}
	// do updates to sheet
	if (m_ongoingDrawCommand) {
		m_updated = m_updated || m_ongoingDrawCommand->append(idxList);
	} else if (getPixel(activeSubSheet, pt) != palIdx) {
		pushCommand(ox::make<DrawCommand>(m_img, m_activeSubsSheetIdx, idxList, palIdx));
	}
}

void TileSheetEditorModel::setSelection(studio::Selection const&sel) noexcept {
	m_selection.emplace(sel);
	m_updated = true;
}

void TileSheetEditorModel::select(ox::Point const&pt) noexcept {
	if (m_selTracker.updateCursorPoint(pt)) {
		setSelection(m_selTracker.selection());
	}
}

void TileSheetEditorModel::completeSelection() noexcept {
	if (m_selTracker.selectionOngoing()) {
		m_selTracker.finishSelection();
		m_selection.emplace(m_selTracker.selection());
		auto&pt = m_selection->b;
		auto const&s = activeSubSheet();
		pt.x = ox::min(s.columns * TileWidth - 1, pt.x);
		pt.y = ox::min(s.rows * TileHeight - 1, pt.y);
	}
}

void TileSheetEditorModel::clearSelection() noexcept {
	m_updated = true;
	m_selTracker.reset();
	m_selection.reset();
}

bool TileSheetEditorModel::updated() const noexcept {
	return m_updated;
}

ox::Error TileSheetEditorModel::markUpdatedCmdId(studio::UndoCommand const*cmd) noexcept {
	m_updated = true;
	const auto cmdId = cmd->commandId();
	if (static_cast<CommandId>(cmdId) == CommandId::PaletteChange) {
		OX_RETURN_ERROR(readObj<Palette>(keelCtx(m_tctx), m_img.defaultPalette).moveTo(m_pal));
		m_palettePage = ox::min<size_t>(m_pal->pages.size(), 0);
		paletteChanged.emit();
	}
	auto tsCmd = dynamic_cast<const TileSheetCommand*>(cmd);
	auto idx = validateSubSheetIdx(m_img, tsCmd->subsheetIdx());
	if (idx != m_activeSubsSheetIdx) {
		setActiveSubsheet(idx);
	}
	return {};
}

ox::Error TileSheetEditorModel::markUpdated() noexcept {
	m_updated = true;
	return {};
}

void TileSheetEditorModel::ackUpdate() noexcept {
	m_updated = false;
}

ox::Error TileSheetEditorModel::saveFile() noexcept {
	return m_sctx.project->writeObj(m_path, m_img, ox::ClawFormat::Organic);
}

bool TileSheetEditorModel::pixelSelected(std::size_t const idx) const noexcept {
	auto const&s = activeSubSheet();
	auto const pt = idxToPt(static_cast<int>(idx), s.columns);
	return m_selection && m_selection->contains(pt);
}

void TileSheetEditorModel::getFillPixels(
		TileSheet::SubSheet const&activeSubSheet,
		ox::Span<bool> pixels,
		ox::Point const&pt,
		int const oldColor) const noexcept {
	auto const idx = ptToIdx(pt, activeSubSheet.columns);
	auto const relIdx = idx % PixelsPerTile;
	if (pixels[relIdx] || getPixel(activeSubSheet, idx) != oldColor) {
		return;
	}
	// mark pixels to update
	pixels[relIdx] = true;
	if (pt.x % TileWidth != 0) {
		auto const leftPt = pt + ox::Point{-1, 0};
		getFillPixels(activeSubSheet, pixels, leftPt, oldColor);
	}
	if (pt.x % TileWidth != TileWidth - 1) {
		auto const rightPt = pt + ox::Point{1, 0};
		getFillPixels(activeSubSheet, pixels, rightPt, oldColor);
	}
	if (pt.y % TileHeight != 0) {
		auto const topPt = pt + ox::Point{0, -1};
		getFillPixels(activeSubSheet, pixels, topPt, oldColor);
	}
	if (pt.y % TileHeight != TileHeight - 1) {
		auto const bottomPt = pt + ox::Point{0, 1};
		getFillPixels(activeSubSheet, pixels, bottomPt, oldColor);
	}
}

void TileSheetEditorModel::pushCommand(studio::UndoCommand *cmd) noexcept {
	std::ignore = m_undoStack.push(ox::UPtr<studio::UndoCommand>{cmd});
	m_ongoingDrawCommand = dynamic_cast<DrawCommand*>(cmd);
	m_updated = true;
}

}
